import warnings

import numpy as np


def _check_ratio(ratio):
    if not (0 <= ratio <= 1):
        raise ValueError(f"ratio must be between 0 and 1, but received {ratio}")


def ratio_split(dataset, *ratios):
    if isinstance(ratios[0], tuple):
        ratios = ratios[0]
    for ratio in ratios:
        _check_ratio(ratio)
    if sum(ratios) != 1:
        warnings.warn("some data may be lost since ratio does not sum up to 1")
    splits = []
    start = 0
    index = np.random.permutation(np.arange(len(dataset)))
    for ratio in ratios:
        end = start + int(ratio * len(dataset))
        splits.append(Subset(dataset, index[start:end]))
        start = end
    return splits


class Dataset:
    def __init__(self, transform=None, target_transform=None):
        self.transform = transform
        self.target_transform = target_transform
        if self.transform is None:
            self.transform = lambda x: x
        if self.target_transform is None:
            self.target_transform = lambda x: x
        self.data = None
        self.label = None

    def __getitem__(self, index):
        if self.label is None:
            return self.transform(self.data[index]), None
        return (
            self.transform(self.data[index]),
            self.target_transform(self.label[index]),
        )

    def __len__(self):
        return len(self.data)


class Subset:
    def __init__(self, dataset, indices):
        self.dataset = dataset
        self.indices = indices

    def __getitem__(self, index):
        return self.dataset[self.indices[index]]

    def __len__(self):
        return len(self.indices)


class DataLoader:
    def __init__(self, dataset, batch_size, shuffle=True):
        self.dataset = dataset
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.iter = 0
        self.max_iter = len(dataset) // batch_size
        self.reset()

    def __iter__(self):
        return self

    def __next__(self):
        i = self.iter
        if i == self.max_iter:
            self.reset()
            raise StopIteration
        batch_size = self.batch_size
        batch_index = self.index[i * batch_size : (i + 1) * batch_size]
        batch = self.dataset[batch_index]
        self.iter += 1
        return batch

    def reset(self):
        self.iter = 0
        if self.shuffle:
            self.index = np.random.permutation(np.arange(len(self.dataset)))
        else:
            self.index = np.arange(len(self.dataset))



import weakref

import numpy as np

import pygrad.functions as F
from pygrad import utils
from pygrad.core import Parameter, as_tuple


class Module:
    def __init__(self):
        self._params = set()
        self.is_train = True

    def __setattr__(self, name, value):
        if isinstance(value, (Parameter, Module)):
            self._params.add(name)
        super(Module, self).__setattr__(name, value)

    def __call__(self, *inputs):
        outputs = as_tuple(self.forward(*inputs))
        self.inputs = [weakref.ref(x) for x in inputs]
        self.outputs = [weakref.ref(y) for y in outputs]
        if len(outputs) > 1:
            return outputs
        return outputs[0]

    def __repr__(self):
        return str(self.__dict__)

    def forward(self, inputs):
        raise NotImplementedError

    def params(self):
        for name in self._params:
            param = self.__dict__[name]
            if isinstance(param, Module):
                yield from param.params()
            else:
                yield param

    def train(self, is_train=True):
        self.is_train = is_train
        for name in self._params:
            param = self.__dict__[name]
            if isinstance(param, Module):
                param.train(is_train)

    def eval(self):
        self.train(False)

    def plot(self, to_file="graph.png", dpi=300):
        try:
            self.inputs
        except AttributeError:
            raise RuntimeError("need to run a forward pass first")
        utils.plot_model(self, to_file, dpi)

    def weight_dict(self):
        weights = {}
        for name in self._params:
            param = self.__dict__[name]
            if isinstance(param, Module):
                weights[name] = param.weight_dict()
            else:
                weights[name] = param.data
        return weights

    def load_weight_dict(self, weights):
        for name, weight in weights.items():
            param = self.__dict__[name]
            if isinstance(param, Module):
                if not isinstance(weight, dict):
                    raise KeyError(
                        "cannot load model due to missing or mismatching keys"
                    )
                param.load_weight_dict(weight)
            else:
                if param.data.shape != weight.shape:
                    raise KeyError(
                        "cannot load model due to missing or mismatching keys"
                    )
                param.data = weight

    def load(self, path):
        weights = np.load(path, allow_pickle=True).item()
        self.load_weight_dict(weights)

    def save(self, path):
        weights = self.weight_dict()
        np.save(path, weights)


class Linear(Module):
    def __init__(self, in_size, out_size, bias=True):
        super(Linear, self).__init__()
        self.W = Parameter(np.random.randn(in_size, out_size), name="W")
        if bias:
            self.b = Parameter(np.zeros(out_size), name="b")
        else:
            self.b = None

    def forward(self, x):
        return F.linear(x, self.W, self.b)


class Dropout(Module):
    def __init__(self, dropout=0.5):
        super(Dropout, self).__init__()
        self.dropout = Parameter(dropout)

    def forward(self, x):
        return F.dropout(x, self.dropout, self.is_train)



import pygrad.data
import pygrad.exceptions
import pygrad.functions
import pygrad.nn
from pygrad.core import Parameter, Variable, as_variable, no_grad, setup_variable

setup_variable()

__version__ = "0.1.0"



import contextlib
import heapq
import warnings
import weakref

import numpy as np

import pygrad


class Config:
    enable_backprop = True


@contextlib.contextmanager
def using_config(name, value):
    prev_value = getattr(Config, name)
    setattr(Config, name, value)
    try:
        yield
    finally:
        setattr(Config, name, prev_value)


def no_grad():
    return using_config("enable_backprop", False)


def as_tuple(x):
    if not isinstance(x, tuple):
        return (x,)
    return x


def as_variable(obj):
    if isinstance(obj, Variable):
        return obj
    return Variable(obj)


class Variable:

    __array_priority__ = 100

    def __init__(self, data, name=None):
        if isinstance(data, Variable):
            data = data.data
        self.data = np.asarray(data)
        data_dtype = self.data.dtype
        if not np.can_cast(data_dtype, np.number):
            raise TypeError(
                f"invalid data type '{data_dtype.type.__name__}' for `data`"
            )
        if name and not isinstance(name, str):
            raise TypeError(f"invalid data type '{type(name).__name__}' for `name`")
        self.name = name
        self.grad = None
        self.creator = None
        self.generation = 0

    def __len__(self):
        return len(self.data)

    def __eq__(self, other):
        if isinstance(other, Variable):
            return np.array_equal(self.data, other.data)
        return False

    def __repr__(self):
        data_string = str(self.data).replace("\n", "\n" + " " * 9)
        if self.name is None:
            return f"Variable({data_string})"
        return f"Variable({data_string}), {self.name}"

    @property
    def shape(self):
        return self.data.shape

    @property
    def ndim(self):
        return self.data.ndim

    @property
    def size(self):
        return self.data.size

    @property
    def dtype(self):
        return self.data.dtype

    @property
    def T(self):
        if self.ndim == 0:
            return self
        return pygrad.functions.transpose(self)

    def transpose(self, *axes):
        if self.ndim == 0:
            return self
        if len(axes) == 0:
            axes = None
        elif len(axes) == 1 and isinstance(axes[0], (tuple, list)) or axes[0] is None:
            axes = axes[0]
        if len(axes) > self.ndim:
            raise pygrad.exceptions.AxisError(axes, self.ndim)
        return pygrad.functions.transpose(self, axes)

    def sum(self, axis=None, keepdims=False):
        try:
            return pygrad.functions.sum(self, axis, keepdims)
        except np.AxisError:
            raise pygrad.exceptions.AxisError(axis, self.ndim)

    def reshape(self, *shape):
        if len(shape) == 1 and isinstance(shape, (tuple, list)):
            shape = shape[0]
        return pygrad.functions.reshape(self, shape)

    def set_creator(self, func):
        self.creator = func
        self.generation = func.generation + 1
        return self

    def clear_grad(self):
        self.grad = None

    def backward(self, retain_grad=False, create_graph=False):
        if self.creator is None:
            raise RuntimeError("backward pass on a root variable")
        if self.data.size != 1:
            warnings.warn("attempting backward pass on a non-scalar variable")
        if self.grad is None:
            self.grad = Variable(np.ones_like(self.data))
        funcs = [self.creator]
        seen_set = set(funcs)
        while funcs:
            f = heapq.heappop(funcs)
            gys = [output().grad for output in f.outputs]
            with using_config("enable_backprop", create_graph):
                gxs = as_tuple(f.backward(*gys))
                for x, gx in zip(f.inputs, gxs):
                    if x.grad is None:
                        x.grad = gx
                    else:
                        x.grad = x.grad + gx
                    if not (x.creator is None or x.creator in seen_set):
                        seen_set.add(x.creator)
                        heapq.heappush(funcs, x.creator)
            if not retain_grad:
                for y in f.outputs:
                    y().clear_grad()


class Parameter(Variable):
    pass


class Function:
    def __call__(self, *inputs):
        inputs = [as_variable(x) for x in inputs]
        xs = [x.data for x in inputs]
        ys = as_tuple(self.forward(*xs))
        outputs = [as_variable(y) for y in ys]
        if Config.enable_backprop:
            self.inputs = inputs
            self.generation = max([x.generation for x in inputs])
            self.outputs = [weakref.ref(output.set_creator(self)) for output in outputs]
        if len(outputs) > 1:
            return outputs
        return outputs[0]

    def __lt__(self, other):
        return self.generation > other.generation

    def forward(self, xs):
        raise NotImplementedError

    def backward(self, gys):
        raise NotImplementedError


class Add(Function):
    def forward(self, x0, x1):
        return x0 + x1

    def backward(self, gy):
        x0, x1 = self.inputs
        gx0 = gy
        gx1 = gy
        if x0.shape != x1.shape:
            gx0 = pygrad.functions.sum_to(gx0, x0.shape)
            gx1 = pygrad.functions.sum_to(gx1, x1.shape)
        return gx0, gx1


def add(x0, x1):
    return Add()(x0, x1)


class Neg(Function):
    def forward(self, x):
        return -x

    def backward(self, gy):
        return -gy


def neg(x):
    return Neg()(x)


class Sub(Function):
    def forward(self, x0, x1):
        return x0 - x1

    def backward(self, gy):
        x0, x1 = self.inputs
        gx0 = gy
        gx1 = -gy
        if x0.shape != x1.shape:
            gx0 = pygrad.functions.sum_to(gx0, x0.shape)
            gx1 = pygrad.functions.sum_to(gx1, x1.shape)
        return gx0, gx1


def sub(x0, x1):
    return Sub()(x0, x1)


def rsub(x0, x1):
    return Sub()(x1, x0)


class Mul(Function):
    def forward(self, x0, x1):
        return x0 * x1

    def backward(self, gy):
        x0, x1 = self.inputs
        gx0 = gy * x1
        gx1 = gy * x0
        if x0.shape != x1.shape:
            gx0 = pygrad.functions.sum_to(gx0, x0.shape)
            gx1 = pygrad.functions.sum_to(gx1, x1.shape)
        return gx0, gx1


def mul(x0, x1):
    return Mul()(x0, x1)


class Div(Function):
    def forward(self, x0, x1):
        if 0 in x1:
            raise ZeroDivisionError("division by variable containing zero")
        return x0 / x1

    def backward(self, gy):
        x0, x1 = self.inputs
        gx0 = gy / x1
        gx1 = -gy * x0 / (x1 ** 2)
        if x0.shape != x1.shape:
            gx0 = pygrad.functions.sum_to(gx0, x0.shape)
            gx1 = pygrad.functions.sum_to(gx1, x1.shape)
        return gx0, gx1


def div(x0, x1):
    return Div()(x0, x1)


def rdiv(x0, x1):
    return Div()(x1, x0)


class Pow(Function):
    def __init__(self, c):
        self.c = c

    def forward(self, x):
        return x ** self.c

    def backward(self, gy):
        x = self.inputs[0]
        return gy * self.c * x ** (self.c - 1)


def pow(x, c):
    return Pow(c)(x)


def setup_variable():
    Variable.__neg__ = neg
    Variable.__add__ = add
    Variable.__radd__ = add
    Variable.__sub__ = sub
    Variable.__rsub__ = rsub
    Variable.__mul__ = mul
    Variable.__rmul__ = mul
    Variable.__truediv__ = div
    Variable.__rtruediv__ = rdiv
    Variable.__pow__ = pow
    Variable.__getitem__ = pygrad.functions.get_item



import math

import numpy as np


def _check_lr(lr):
    if 0 > lr:
        raise ValueError(f"expected learning rate to be larger than 0, but got {lr}")


class Optimizer:
    def __init__(self, params):
        self.hooks = []
        self.params = tuple(params)

    def step(self):
        for hook in self.hooks:
            hook(params)
        for param in self.params:
            self.step_one(param)

    def step_one(self, param):
        raise NotImplementedError

    def zero_grad(self):
        for param in self.params:
            param.clear_grad()

    def add_hook(self, hook):
        self.hooks.append(hook)


class SGD(Optimizer):
    def __init__(self, params, lr=1e-2, momentum=0.9):
        _check_lr(lr)
        super(SGD, self).__init__(params)
        self.vs = {}
        self.lr = lr
        self.momentum = momentum

    def step_one(self, param):
        v_key = id(param)
        if v_key not in self.vs:
            self.vs[v_key] = np.zeros_like(param.data)
        v = self.vs[v_key]
        v *= self.momentum
        v -= self.lr * param.grad.data
        param.data += v


class AdaGrad(Optimizer):
    def __init__(self, params, lr=1e-2, eps=1e-10):
        _check_lr(lr)
        super(AdaGrad, self).__init__(params)
        self.gs = {}
        self.lr = lr
        self.eps = eps

    def step_one(self, param):
        g_key = id(param)
        if g_key not in self.gs:
            self.gs[g_key] = np.zeros_like(param.data)
        grad = param.grad.data
        g = self.gs[g_key]
        g += grad * grad
        param.data -= self.lr * grad / (np.sqrt(g + self.eps))


class AdaDelta(Optimizer):
    def __init__(self, params, lr=1.0, rho=0.95, eps=1e-6):
        _check_lr(lr)
        super(AdaDelta, self).__init__(params)
        self.gs = {}
        self.dxhs = {}
        self.lr = lr
        self.rho = rho
        self.eps = eps

    def step_one(self, param):
        key = id(param)
        if key not in self.gs:
            self.gs[key] = np.zeros_like(param.data)
            self.dxhs[key] = np.zeros_like(param.data)
        rho = self.rho
        eps = self.eps
        grad = param.grad.data
        g = self.gs[key]
        g *= rho
        g += (1 - rho) * grad * grad
        dxh = self.dxhs[key]
        dx = np.sqrt((dxh + eps) / (g + eps)) * grad
        dxh *= rho
        dxh += (1 - rho) * dx * dx
        param.data -= self.lr * dx


class Adam(Optimizer):
    def __init__(self, params, lr=1e-3, beta_1=0.9, beta_2=0.999, eps=1e-7):
        _check_lr(lr)
        super(Adam, self).__init__(params)
        self.lr = lr
        self.b1 = beta_1
        self.b2 = beta_2
        self.eps = eps
        self.t = 0
        self.ms = {}
        self.vs = {}

    def step(self, *args, **kwargs):
        self.t += 1
        super(Adam, self).step(*args, **kwargs)

    @property
    def alpha(self):
        denom = 1.0 - math.pow(self.b1, self.t)
        numer = 1.0 - math.pow(self.b2, self.t)
        return self.lr * math.sqrt(numer) / denom

    def step_one(self, param):
        key = id(param)
        if key not in self.ms:
            self.ms[key] = np.zeros_like(param.data)
            self.vs[key] = np.zeros_like(param.data)
        b1 = self.b1
        b2 = self.b2
        grad = param.grad.data
        m = self.ms[key]
        v = self.vs[key]
        m += (1 - b1) * (grad - m)
        v += (1 - b2) * (grad * grad - v)
        param.data -= self.alpha * m / (np.sqrt(v) + self.eps)




import os
import subprocess
import sys
import warnings

import matplotlib.pyplot as plt
import numpy as np

from pygrad.core import Parameter


def numerical_grad(f, x, eps=1e-4):
    return (f(x + eps) - f(x - eps)) / (2 * eps)


def handle_shape(func):
    def wrapper(x, shape):
        if x.shape == shape:
            return x
        return func(x, shape)

    return wrapper


def set_module(module):
    def decorator(func):
        if module is not None:
            func.__module__ = module
        return func

    return decorator


def _sum_to(x, shape):
    ndim = len(shape)
    lead = x.ndim - ndim
    if lead < 0:
        raise ValueError("`shape` has more dimensions than provided variable")
    lead_axis = tuple(range(lead))
    axis = tuple([i + lead for i, sx in enumerate(shape) if sx == 1])
    y = x.sum(lead_axis + axis, keepdims=True)
    if lead > 0:
        y = y.squeeze(lead_axis)
    if y.shape != shape:
        raise RuntimeError(f"variable cannot be summed to shape {shape}")
    return y


def _log_sum_exp(x, axis):
    m = x.max(axis=axis, keepdims=True)
    y = np.exp(x - m).sum(axis=axis, keepdims=True)
    return np.add(m, np.log(y))


def write_dot_graph(model, dpi):
    nodes = {}
    edges = {}
    funcs = [output().creator for output in model.outputs]
    seen_set = set(funcs)
    while funcs:
        f = funcs.pop()
        node, edge = make_dot(f)
        nodes.update(node)
        edges.update(edge)
        for input_ in f.inputs:
            creator = input_.creator
            if creator and creator not in seen_set:
                seen_set.add(creator)
                funcs.append(creator)
    txt = ""
    for node_id, properties in nodes.items():
        txt += (
            f'{node_id} [label="{properties[0]} | {{ in | out }} | '
            f'{{ {properties[1]} | {properties[2]} }}"]\n'
        )
    for root, child in edges.items():
        txt += f"{root} -> {child}\n"
        if root not in nodes:
            txt += f'{root} [label="Variable"]\n'
    return (
        f'digraph g \n{{edge [fontname = "helvetica"];'
        f'graph [fontname = "helvetica", dpi={dpi}];'
        f'node [fontname = "helvetica", shape=Mrecord];\n{txt}}}'
    )


def make_dot(f):
    f_inputs = [input_ for input_ in f.inputs if not isinstance(input_, Parameter)]
    input_shapes = [input_.shape for input_ in f_inputs]
    output_shapes = [output().shape for output in f.outputs]
    name = f.__class__.__name__
    id_ = id(f)
    node = {id_: [name, input_shapes, output_shapes]}
    creator_ids = [id(input_.creator) for input_ in f_inputs]
    edge = {creator_id: id_ for creator_id in creator_ids}
    return node, edge


def _check_graphviz():
    try:
        subprocess.run("dot -V", shell=True, check=True)
    except subprocess.CalledProcessError:
        raise ImportError(
            "please install graphviz (https://graphviz.gitlab.io/download/)"
        )


def plot_model(model, to_file="graph.png", dpi=300):
    _check_graphviz()
    graph = write_dot_graph(model, dpi)
    tmp_dir = os.path.join(os.path.expanduser("~"), ".pygrad")
    if not os.path.exists(tmp_dir):
        os.mkdir(tmp_dir)
    graph_file = os.path.join(tmp_dir, "graph.dot")
    with open(graph_file, "w+") as f:
        f.write(graph)
    extension = os.path.splitext(to_file)[-1][1:]
    save_path = os.path.join(os.getcwd(), to_file)
    cmd = f"dot {graph_file} -T {extension} -o {save_path}"
    try:
        subprocess.run(cmd, shell=True, check=True)
    except subprocess.CalledProcessError:
        raise FileNotFoundError(f"no such file or directory {save_path}")
    plt.axis("off")
    try:
        from IPython import display

        display.set_matplotlib_formats("retina")
        return diplay.Image(filename=to_file)
    except NameError:
        import matplotlib.image as mpimg

        img = mpimg.imread(to_file)
        plt.imshow(img)
        plt.show()



from setuptools import find_packages, setup

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="pygrad",
    version="0.1.0",
    author="Jake Tae",
    author_email="jaesungtae@gmail.com",
    description="A pure Python autograd library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jaketae/pygrad",
    packages=find_packages(exclude=["docs", "tests"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=["numpy", "matplotlib"],
)



import numpy as np

from pygrad.core import Function, as_variable
from pygrad.utils import _log_sum_exp, _sum_to, handle_shape


class Exp(Function):
    def forward(self, x):
        return np.exp(x)

    def backward(self, gy):
        return gy * exp(self.inputs[0])


def exp(x):
    return Exp()(x)


class Square(Function):
    def forward(self, x):
        return x ** 2

    def backward(self, gy):
        return 2 * gy * self.inputs[0]


def square(x):
    return Square()(x)


class Sin(Function):
    def forward(self, x):
        return np.sin(x)

    def backward(self, gy):
        return gy * cos(self.inputs[0])


def sin(x):
    return Sin()(x)


class Cos(Function):
    def forward(self, x):
        return np.cos(x)

    def backward(self, gy):
        return -gy * sin(self.inputs[0])


def cos(x):
    return Cos()(x)


class Tanh(Function):
    def forward(self, x):
        return np.tanh(x)

    def backward(self, gy):
        y = self.outputs[0]()
        return gy * (1 - y * y)


def tanh(x):
    return Tanh()(x)


class Reshape(Function):
    def __init__(self, shape):
        self.shape = shape

    def forward(self, x):
        self.x_shape = x.shape
        return x.reshape(self.shape)

    def backward(self, gy):
        return reshape(gy, self.x_shape)


@handle_shape
def reshape(x, shape):
    return Reshape(shape)(x)


class Transpose(Function):
    def __init__(self, axes):
        self.axes = axes

    def forward(self, x):
        return x.transpose(self.axes)

    def backward(self, gy):
        if self.axes is None:
            return transpose(gy)
        inv_axes = np.argsort(self.axes)
        return transpose(gy, inv_axes)


def transpose(x, axes=None):
    return Transpose(axes)(x)


class Sum(Function):
    def __init__(self, axis, keepdims):
        self.axis = axis
        self.keepdims = keepdims

    def forward(self, x):
        self.x_shape = x.shape
        return x.sum(axis=self.axis, keepdims=self.keepdims)

    def backward(self, gy):
        return broadcast_to(gy, self.x_shape)


def sum(x, axis=None, keepdims=False):
    return Sum(axis, keepdims)(x)


class BroadcastTo(Function):
    def __init__(self, shape):
        self.shape = shape

    def forward(self, x):
        self.x_shape = x.shape
        return np.broadcast_to(x, self.shape)

    def backward(self, gy):
        return sum_to(gy, self.x_shape)


@handle_shape
def broadcast_to(x, shape):
    return BroadcastTo(shape)(x)


class SumTo(Function):
    def __init__(self, shape):
        self.shape = shape

    def forward(self, x):
        self.x_shape = x.shape
        return _sum_to(x, self.shape)

    def backward(self, gy):
        return broadcast_to(gy, self.x_shape)


@handle_shape
def sum_to(x, shape):
    return SumTo(shape)(x)


class MatMul(Function):
    def forward(self, x, W):
        return x.dot(W)

    def backward(self, gy):
        x, W = self.inputs
        gx = matmul(gy, W.T)
        gW = matmul(x.T, gy)
        return gx, gW


def matmul(x, W):
    return MatMul()(x, W)


class MeanSquaredError(Function):
    def forward(self, x0, x1):
        diff = x1 - x0
        return np.average(diff ** 2)

    def backward(self, gy):
        x0, x1 = self.inputs
        diff = x1 - x0
        gx0 = 2 * gy * diff / len(diff)
        gx1 = -gx0
        return gx0, gx1


def mean_squared_error(x0, x1):
    return MeanSquaredError()(x0, x1)


class Linear(Function):
    def forward(self, x, W, b):
        y = x.dot(W)
        if b is not None:
            y += b
        return y

    def backward(self, gy):
        x, W, b = self.inputs
        gb = None
        if b.data is not None:
            gb = sum_to(gy, b.shape)
        gx = matmul(gy, W.T)
        gW = matmul(x.T, gy)
        return gx, gW, gb


def linear(x, W, b=None):
    return Linear()(x, W, b)


class Sigmoid(Function):
    def forward(self, x):
        return 1 / (1 + np.exp(-x))

    def backward(self, gy):
        y = self.outputs[0]()
        return gy * y * (1 - y)


def sigmoid(x):
    return Sigmoid()(x)


class ReLU(Function):
    def forward(self, x):
        y = x.copy()
        y[y < 0] = 0
        return y

    def backward(self, gy):
        x = self.inputs[0]
        mask = x.data > 0
        return gy * mask


def relu(x):
    return ReLU()(x)


class LeakyReLU(Function):
    def __init__(self, slope):
        self.slope = slope

    def forward(self, x):
        y = x.copy()
        y[x < 0] *= self.slope
        return y

    def backward(self, gy):
        x = self.inputs[0]
        mask = x.data > 0
        mask[mask <= 0] = self.slope
        return gy * mask


def leaky_relu(x, slope=0.2):
    return LeakyReLU(slope)(x)


class GetItem(Function):
    def __init__(self, slice):
        self.slice = slice

    def forward(self, x):
        return x[self.slice]

    def backward(self, gy):
        x = self.inputs[0]
        gx = np.zeros_like(x.data)
        np.add.at(gx, self.slice, gy.data)
        return gx


def get_item(x, slice):
    return GetItem(slice)(x)


class Softmax(Function):
    def __init__(self, axis):
        self.axis = axis

    def forward(self, x):
        y = np.exp(x - x.max(axis=self.axis, keepdims=True))
        return y / y.sum(axis=self.axis, keepdims=True)

    def backward(self, gy):
        y = self.outputs[0]()
        gx = y * gy
        gx_sum = gx.sum(axis=self.axis, keepdims=True)
        return gx - y * gx_sum


def softmax(x, axis=-1):
    return Softmax(axis)(x)


class Log(Function):
    def forward(self, x):
        return np.log(x)

    def backward(self, gy):
        x = self.inputs[0]
        return gy / x


def log(x):
    return Log()(x)


class LogSoftMax(Function):
    def __init__(self, axis):
        self.axis = axis

    def forward(self, x):
        log_z = _log_sum_exp(x, self.axis)
        return x - log_z

    def backward(self, gy):
        y = self.outputs[0]()
        return gy - exp(y) * gy.sum(axis=self.axis, keepdims=True)


def log_softmax(x, axis=-1):
    return LogSoftMax(axis)(x)


class SoftmaxCrossEntropy(Function):
    def __init__(self, axis):
        self.axis = axis

    def forward(self, x, t):
        if x.ndim != 2:
            raise ValueError(
                f"expected `x` predicted values to be 2D, "
                f"but received {x.ndim}D variable instead"
            )
        if t.ndim != 1:
            raise ValueError(
                f"expected `t` label to be 1D, "
                f"but received {t.ndim}D variable instead"
            )
        num_labels = len(t)
        num_samples = x.shape[0]
        if num_labels != num_samples:
            raise ValueError(
                f"expected `x` and `t` to be of same length, "
                f"but received {num_samples} samples, {num_labels} labels"
            )
        log_z = _log_sum_exp(x, self.axis)
        log_p = x - log_z
        log_p = log_p[:, t.ravel()]
        return -np.average(log_p)

    def backward(self, gy):
        x, t = self.inputs
        batch_size, num_class = x.shape
        y = softmax(x)
        t_onehot = np.eye(num_class)[t.data]
        return (y - t_onehot) * gy / batch_size


def softmax_cross_entropy(x, t, axis=-1):
    return SoftmaxCrossEntropy(axis)(x, t)


class Dropout(Function):
    def __init__(self, dropout, train):
        self.train = train
        self.dropout = dropout.data

    def forward(self, x):
        dropout = self.dropout
        if not (0 <= dropout <= 1):
            raise ValueError("`dropout` must be between 0 and 1")
        if self.train:
            self.mask = np.random.randn(*x.shape) > dropout
            self.scale = np.array(1 - dropout)
            return x * self.mask / self.scale
        return x

    def backward(self, gy):
        if self.train:
            return gy * self.mask / self.scale
        return gy


def dropout(x, dropout, train):
    return Dropout(dropout, train)(x)




from pygrad.utils import set_module


@set_module("pygrad")
class AxisError(IndexError, ValueError):
    def __init__(self, axis, ndim):
        super(AxisError, self).__init__(
            f"axis {axis} is out of bounds for variable of dimension {ndim}"
        )
